# Copyright (c) 2016 FUJITSU LIMITED
# Copyright (c) 2012 EMC Corporation.
# Copyright (c) 2012 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""
FibreChannel Cinder Volume driver for Fujitsu ETERNUS DX S3 series.
"""
from oslo_log import log as logging
import six

from cinder.i18n import _LI
from cinder.volume import driver
from cinder.volume.drivers.fujitsu import eternus_dx_common
from cinder.zonemanager import utils as fczm_utils

LOG = logging.getLogger(__name__)


class FJDXFCDriver(driver.FibreChannelDriver, driver.ReplicaV2VD):
    """FC Cinder Volume Driver for Fujitsu ETERNUS DX S3 series."""

    def __init__(self, *args, **kwargs):
        super(FJDXFCDriver, self).__init__(*args, **kwargs)
        self.common = eternus_dx_common.FJDXCommon(
            'fc',
            configuration=self.configuration)
        self.VERSION = self.common.VERSION

    def check_for_setup_error(self):
        pass

    def create_volume(self, volume):
        """Create volume."""
        LOG.info(_LI('create_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        model_update = self.common.create_volume(volume)

        LOG.info(_LI('create_volume, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def create_volume_from_snapshot(self, volume, snapshot):
        """Creates a volume from a snapshot."""
        LOG.info(_LI('create_volume_from_snapshot, '
                     'volume id: %(vid)s, snap id: %(sid)s, Enter method.'),
                 {'vid': volume['id'], 'sid': snapshot['id']})

        model_update = (
            self.common.create_volume_from_snapshot(volume, snapshot))

        LOG.info(_LI('create_volume_from_snapshot, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def create_cloned_volume(self, volume, src_vref):
        """Create cloned volume."""
        LOG.info(_LI('create_cloned_volume, '
                     'target volume id: %(tid)s, '
                     'source volume id: %(sid)s, Enter method.'),
                 {'tid': volume['id'], 'sid': src_vref['id']})

        model_update = (
            self.common.create_cloned_volume(volume, src_vref))

        LOG.info(_LI('create_cloned_volume, '
                     'info: %s, Exit method.'), model_update['metadata'])
        return model_update

    def delete_volume(self, volume):
        """Delete volume on ETERNUS."""
        LOG.info(_LI('delete_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        vol_exist = self.common.delete_volume(volume)

        LOG.info(_LI('delete_volume, '
                     'delete: %s, Exit method.'), vol_exist)

    def create_snapshot(self, snapshot):
        """Creates a snapshot."""
        LOG.info(_LI('create_snapshot, '
                     'snap id: %(sid)s, volume id: %(vid)s, Enter method.'),
                 {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        location, metadata = self.common.create_snapshot(snapshot)

        LOG.info(_LI('create_snapshot, info: %s, Exit method.'), metadata)
        return {'provider_location': six.text_type(location)}

    def delete_snapshot(self, snapshot):
        """Deletes a snapshot."""
        LOG.info(_LI('delete_snapshot, '
                     'snap id: %(sid)s, volume id: %(vid)s, Enter method.'),
                 {'sid': snapshot['id'], 'vid': snapshot['volume_id']})

        vol_exist = self.common.delete_snapshot(snapshot)

        LOG.info(_LI('delete_snapshot, '
                     'delete: %s, Exit method.'), vol_exist)

    def ensure_export(self, context, volume):
        """Driver entry point to get the export info for an existing volume."""
        pass

    def create_export(self, context, volume, connector):
        """Driver entry point to get the export info for a new volume."""
        pass

    def remove_export(self, context, volume):
        """Driver entry point to remove an export for a volume."""
        pass

    @fczm_utils.AddFCZone
    def initialize_connection(self, volume, connector):
        """Allow connection to connector and return connection info."""
        LOG.info(_LI('initialize_connection, volume id: %(vid)s, '
                     'wwpns: %(wwpns)s, Enter method.'),
                 {'vid': volume['id'], 'wwpns': connector['wwpns']})

        info = self.common.initialize_connection(volume, connector)

        data = info['data']
        init_tgt_map = (
            self.common.build_fc_init_tgt_map(connector, data['target_wwn']))
        data['initiator_target_map'] = init_tgt_map

        info['data'] = data
        LOG.info(_LI('initialize_connection, '
                     'info: %s, Exit method.'), info)
        return info

    @fczm_utils.RemoveFCZone
    def terminate_connection(self, volume, connector, **kwargs):
        """Disallow connection from connector."""
        LOG.info(_LI('terminate_connection, volume id: %(vid)s, '
                     'wwpns: %(wwpns)s, Enter method.'),
                 {'vid': volume['id'], 'wwpns': connector['wwpns']})

        map_num = self.common.terminate_connection(volume, connector)
        info = {'driver_volume_type': 'fibre_channel',
                'data': {}}
        if map_num == 0:
            # No more volumes attached to the host.
            init_tgt_map = self.common.build_fc_init_tgt_map(connector)
            info['data'] = {'initiator_target_map': init_tgt_map}

        LOG.info(_LI('terminate_connection, '
                     'map_num: %(map_num)s, info: %(info)s, Exit method'),
                 {'map_num': map_num, 'info': info})
        return info

    def get_volume_stats(self, refresh=False):
        """Get volume stats."""
        LOG.debug('get_volume_stats, refresh: %s, Enter method.', refresh)

        pool_name = None
        if refresh is True:
            data, pool_name = self.common.update_volume_stats()
            backend_name = self.configuration.safe_get('volume_backend_name')
            data['volume_backend_name'] = backend_name or 'FJDXFCDriver'
            data['storage_protocol'] = 'FC'
            self._stats = data

        LOG.debug('get_volume_stats, '
                  'pool name: %s, Exit method.', pool_name)
        return self._stats

    def extend_volume(self, volume, new_size):
        """Extend volume."""
        LOG.info(_LI('extend_volume, '
                     'volume id: %s, Enter method.'), volume['id'])

        used_pool_name = self.common.extend_volume(volume, new_size)

        LOG.info(_LI('extend_volume, '
                     'used pool name: %s, Exit method.'), used_pool_name)

    def manage_existing(self, volume, existing_ref):
        """Bring an existing volume under Cinder management."""
        LOG.info(_LI('manage_existing, volume id: %(vid)s, '
                     'id or name: %(ref)s, Enter method.'),
                 {'vid': volume['id'], 'ref': existing_ref})

        info = self.common.manage_existing(volume, existing_ref)

        LOG.info(_LI('manage_existing, '
                     'info: %s, Exit method.'), info['metadata'])
        return info

    def manage_existing_get_size(self, volume, existing_ref):
        """Return size of volume to be managed by manage existing."""
        LOG.info(_LI('manage_existing_get_size, volume id: %(vid)s, '
                     'id or name: %(ref)s, Enter method.'),
                 {'vid': volume['id'], 'ref': existing_ref})

        vol_size = self.common.manage_existing_get_size(volume, existing_ref)

        LOG.info(_LI('manage_existing_get_size, '
                     'volume size: %s, Exit method.'), vol_size)
        return vol_size

    def unmanage(self, volume):
        """Removes the specified volume from Cinder management."""
        pass

    def update_migrated_volume(self, ctxt, volume, new_volume,
                               original_volume_status):
        """Update migrated volume."""
        LOG.info(_LI('update_migrated_volume, '
                     'source volume id: %(s_id)s, '
                     'target volume id: %(t_id)s, Enter method.'),
                 {'s_id': volume['id'], 't_id': new_volume['id']})

        model_update = self.common.update_migrated_volume(
            ctxt, volume, new_volume)

        LOG.info(_LI('update_migrated_volume, '
                     'target volume meta: %s, Exit method.'), model_update)

        return model_update

    def get_replication_updates(self, context):
        """Get replication updates."""
        return []

    def replication_enable(self, context, volume):
        """Enable replication on a replication capable volume."""
        raise NotImplementedError()

    def replication_disable(self, context, volume):
        """Disable replication on the specified volume."""
        raise NotImplementedError()

    def replication_failover(self, context, volume, secondary):
        """Force failover to a secondary replication target."""
        raise NotImplementedError()

    def list_replication_targets(self, context, vref):
        """Provide a means to obtain replication targets for a volume."""
        raise NotImplementedError()
