# Copyright (c) 2016 FUJITSU LIMITED
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

import functools

from cinder import context
from cinder import exception
from cinder.i18n import _
from cinder.volume import volume_types
from oslo_concurrency import lockutils
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import units

LOG = logging.getLogger(__name__)
CONF = cfg.CONF

STD_VOLUME_COUNT_MAX = 128
POOL_TYPE_list = [
    'RAID',
    'TPP',
]

FJ_ETERNUS_DX_POOL_opts = [
    cfg.StrOpt('fujitsu_scheduler',
               default='default',
               help='scheduling system for selecting pool to create volume'),
]

CONF.register_opts(FJ_ETERNUS_DX_POOL_opts)


def FJDXLockutils(name, lock_file_prefix, external=False, lock_path=None):
    """Lockutils for Fujitsu ETERNUS DX."""
    def decorator(func):
        def wrapper(self, *args, **kwargs):
            lockname = 'ETERNUS_DX-' + name + '-' + self.storage_name

            @lockutils.synchronized(lockname, lock_file_prefix,
                                    external, lock_path)
            @functools.wraps(func)
            def caller():
                return func(self, *args, **kwargs)
            return caller()
        return wrapper
    return decorator


def get_metadata(volume):
    """Get metadata using volume information."""
    LOG.debug('get_metadata, volume id: %s.',
              volume['id'])

    d_metadata = {}

    metadata = volume.get('volume_metadata')
    if metadata:
        for data in metadata:
            d_metadata[data['key']] = data['value']

    LOG.debug('get_metadata, metadata is: %s.',
              d_metadata)

    return d_metadata


def get_extra_specs(volume, key=None, default=None):
    """Get extra specs from volume.

    If key is none, return all extra specs.
    """
    LOG.debug('get_extra_specs, '
              'volume id: %(v_id)s, '
              'key: %(key)s, '
              'default: %(default)s.',
              {'v_id': volume['id'],
               'key': key,
               'default': default})

    extra_specs = None
    ret = default

    volume_type_id = volume.get('volume_type_id')
    if volume_type_id:
        ctxt = context.get_admin_context()
        volume_type = volume_types.get_volume_type(ctxt, volume_type_id)
        extra_specs = volume_type.get('extra_specs')

    if extra_specs:
        if not key:
            ret = extra_specs
        else:
            ret = extra_specs.get(key, default)

    LOG.debug('get_extra_specs: %s.', extra_specs)
    return ret


class FJDXPoolManager(object):
    """Pool Manager for ETERNUS OpenStack Cinder Volume Driver."""

    def __init__(self, configuration, replication_backend=None):
        scheduler_dic = {
            'default': self._skd_default,
            'round-robin': self._skd_round_robin,
            'capacity-first': self._skd_capacity_first,
        }

        self.configuration = configuration
        self.configuration.append_config_values(FJ_ETERNUS_DX_POOL_opts)
        self.pools = []
        self.pool_idx = -1
        self.pool_num = 0

        LOG.debug('replication_backend_name: %s.', replication_backend)
        if replication_backend is not None:
            self.pools_rep = []
            self.pool_idx_rep = -1
            self.pool_num_rep = 0

        self.max_over_subscription_ratio = (
            self.configuration.max_over_subscription_ratio)

        self.scheduler = (
            scheduler_dic.get(self.configuration.fujitsu_scheduler))

        if not self.scheduler:
            msg = (_('Invalid scheduler was specified : %s.')
                   % self.configuration.fujitsu_scheduler)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

    def update_pools(self, pools, remote=False):
        """Update pools."""
        LOG.debug('update_pools, pools: %(pools)s, '
                  'remote: %(remote)s.',
                  {'pools': pools, 'remote': remote})

        if remote:
            self.pools_rep = list(pools)
            self.pool_num_rep = len(pools)
            if self.pool_num_rep < self.pool_idx_rep:
                self.pool_idx_rep = 0
            return

        self.pools = list(pools)
        self.pool_num = len(pools)

        if self.pool_num < self.pool_idx:
            self.pool_idx = 0

        total_capacity_gb = 0
        free_capacity_gb = 0
        thin_provisioning_support = False
        thick_provisioning_support = False

        for pool in pools:
            if pool['type'] == 'TPP':
                thin_provisioning_support = True
            else:  # pool['type'] == 'RAID':
                thick_provisioning_support = True

            total_capacity_gb += pool['total_capacity_gb']
            free_capacity_gb += pool['free_capacity_gb']

        # Provide more informations.
        pools_info = {
            'total_capacity_gb': total_capacity_gb,
            'free_capacity_gb': free_capacity_gb,
            'thin_provisioning_support': thin_provisioning_support,
            'thick_provisioning_support': thick_provisioning_support,
        }

        LOG.debug('update_pools, '
                  'pools: %(pools)s, '
                  'pools info: %(info)s.',
                  {'pools': pools, 'info': pools_info})
        return pools_info

    def create_pool_info(self, pool_instance, volume_count, pool_type):
        """Create pool information from pool instance."""
        LOG.debug('create_pool_info, pool_instance: %(pool)s, '
                  'volume_count: %(volcount)s, pool_type: %(ptype)s.',
                  {'pool': pool_instance,
                   'volcount': volume_count, 'ptype': pool_type})

        if pool_type not in POOL_TYPE_list:
            msg = (_('Invalid pool type was specified : %s.') % pool_type)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        total_gb = pool_instance['TotalManagedSpace'] / units.Gi
        free_gb = pool_instance['RemainingManagedSpace'] / units.Gi
        pool = {
            'name': pool_instance['ElementName'],
            'path': pool_instance.path,
            'total_capacity_gb': total_gb,
            'free_capacity_gb': free_gb,
            'type': pool_type,
            'volume_count': volume_count,
        }

        LOG.debug('create_pool_info, pool: %s.', pool)
        return pool

    def consume_poolspace(self, volume_size_gb, use_tpp=True, remote=False):
        """Consume pool space for volume in storage pool."""
        LOG.debug('consume_poolspace, size: %(size)s, '
                  'remote: %(remote)s.',
                  {'size': volume_size_gb, 'remote': remote})

        pools = []
        valid = False

        if not remote:
            tmplist = self.pools
        else:
            tmplist = self.pools_rep

        max_over_subscription_ratio = self.max_over_subscription_ratio
        LOG.debug('consume_poolspace, remote: %(remote)s, '
                  'max_over_subscription_ratio: %(value)s.',
                  {'remote': remote, 'value': max_over_subscription_ratio})

        # pools.append(None) to keep the pool size and
        # index the same as self.pools.
        for pool in tmplist:
            if pool['type'] == 'RAID':
                if volume_size_gb <= pool['free_capacity_gb']:
                    if pool['volume_count'] < STD_VOLUME_COUNT_MAX:
                        pools.append(pool)
                        valid = True
                    else:
                        pools.append(None)
                else:
                    pools.append(None)

            else:  # pool['type'] == 'TPP'
                if not use_tpp:
                    pools.append(None)
                    continue

                max_capacity = pool['total_capacity_gb'] * (
                    max_over_subscription_ratio)
                provisioned_capacity = pool['total_capacity_gb'] - (
                    pool['free_capacity_gb'])
                if ((volume_size_gb + provisioned_capacity) <= max_capacity):
                    pools.append(pool)
                    valid = True
                else:
                    pools.append(None)

        if not valid:
            msg = (_('No free space in all storage pool for created volume, '
                     'necessary space : %(volume_size)s.')
                   % {'volume_size': volume_size_gb})
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        # The size and index of pools is the same as self.pools.
        if not remote:
            self.pool_idx = self.scheduler(pools)
            selected_pool = self.pools[self.pool_idx]
        else:
            self.pool_idx_rep = self.scheduler(pools, remote=True)
            selected_pool = self.pools_rep[self.pool_idx_rep]

        LOG.debug('consume_poolspace, selected pool: %s.', selected_pool)
        return selected_pool

    def update_poolspace(self, volume_size_gb, poolpath, mode):
        """Update pool space for volume in storage pool."""
        LOG.debug('update_poolspace, size: %(size)s, '
                  'poolpath: %(ppath)s, mode: %(mode)s.',
                  {'size': volume_size_gb,
                   'ppath': poolpath, 'mode': mode})

        for pool in self.pools:
            if pool['path']['InstanceID'] == poolpath['InstanceID']:
                selected_pool = pool
                break
        else:
            msg = (_('Specified pool is not exist : %s.') % poolpath)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        if mode is 'consume':  # create operation
            selected_pool['free_capacity_gb'] -= volume_size_gb
        elif mode is 'release':  # delete operation
            selected_pool['free_capacity_gb'] += volume_size_gb
        else:
            msg = (_('Specified mode is invalid : %s.') % mode)
            LOG.error(msg)
            raise exception.VolumeBackendAPIException(data=msg)

        LOG.debug('update_poolspace, updated capacity: %s',
                  selected_pool['free_capacity_gb'])

    def _skd_default(self, pools, remote=False):
        """Scheduling System Functions.

        Get the index of the first pool which is not none.
        """
        LOG.debug('_skd_default, remote: %(remote)s.',
                  {'remote': remote})

        idx = 0
        no = -1

        for no, pool in enumerate(pools):
            if pool:
                idx = no
                break

        return idx

    def _skd_round_robin(self, pools, remote=False):
        """Scheduling System Functions.

        Begin at the index which is chosen the last time.
        Get the index of the first pool which is not none.
        """
        LOG.debug('_skd_round_robin, remote: %(remote)s.',
                  {'remote': remote})

        idx = 0
        if not remote:
            offset = self.pool_idx + 1
        else:
            offset = self.pool_idx_rep + 1

        for idx, pool in enumerate(pools[offset:]):
            if pool:
                idx += offset
                break
        else:
            for idx, pool in enumerate(pools[:offset]):
                if pool:
                    break

        return idx

    def _skd_capacity_first(self, pools, remote=False):
        """Scheduling System Functions.

        Get the index of the max capacity pool.
        """
        LOG.debug('_skd_capacity_first, remote: %(remote)s.',
                  {'remote': remote})

        idx = 0
        max_capacity = 0
        no = -1

        for no, pool in enumerate(pools):
            if pool:
                capacity = pool['free_capacity_gb']
                if max_capacity < capacity:
                    idx = no
                    max_capacity = capacity

        return idx
